from __future__ import annotations

import time
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, cast
from urllib.parse import urljoin

import numpy as np
import numpy.typing as npt
import requests
import requests.adapters
from typing_extensions import override

if TYPE_CHECKING:
    import stim
from deltakit_explorer._api._api_client import APIClient, APIEndpoints
from deltakit_explorer._api._auth import (
    get_token,
    https_verification_disabled,
    set_token,
)
from deltakit_explorer._utils._logging import Logging
from deltakit_explorer.enums._api_enums import DataFormat
from deltakit_explorer.types._exceptions import ServerException
from deltakit_explorer.types._experiment_types import QECExperimentDefinition
from deltakit_explorer.types._types import (
    BinaryDataType,
    DataString,
    Decoder,
    DecodingResult,
    DetectionEvents,
    LeakageFlags,
    Measurements,
    NoiseModel,
    ObservableFlips,
    QubitCoordinateToDetectorMapping,
)


class JobStatus(Enum):
    """Job statuses."""

    SUBMITTED = "SUBMITTED"
    IN_PROGRESS = "IN_PROGRESS"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"


@dataclass
class Job:
    """Job record, returned by the server."""

    status: str
    request_id: str
    type: APIEndpoints
    error: str | None = None
    workload: dict = field(default_factory=dict)
    result: dict = field(default_factory=dict)

    def raise_on_error(self):
        if self.error is not None:
            raise ServerException(self.error)


class APIv2Client(APIClient):
    """
    APIv2 implementation replaces long-living requests with job-based approach.
    Client library submits the job and polls the server to obtain the result of
    computations. This makes an HTTP request to the server a lightweight
    operation, completed within milliseconds time.
    """

    STATUS_CHECK_DELAY = 2

    def __init__(self, base_url: str, request_timeout: int = 100):
        """
        Constructor of the APIv2 client instance.

        Args:
            base_url (str): URL of the service endpoint.
            request_timeout (int):
                Timeout (seconds) on the transport level.
                Matches Cloudflare 100s timeout by default.
        """
        super().__init__(base_url, request_timeout)
        self.add_task_endpoint = f"{self.base_url}/api/v2/tasks/add/"
        self.status_endpoint = f"{self.base_url}/api/v2/tasks/get/"
        self.delete_endpoint = f"{self.base_url}/api/v2/tasks/kill/"
        self.request_timeout = request_timeout

        self.auth_headers: dict[str, str] = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {get_token()}",
        }
        self._request_session = requests.Session()
        retries = requests.adapters.Retry(
            total=3,
            backoff_factor=0.5,
            status_forcelist=[500, 502, 503, 504],
        )
        self._request_session.mount(
            "http://",
            requests.adapters.HTTPAdapter(max_retries=retries),
        )
        self._request_session.mount(
            "https://",
            requests.adapters.HTTPAdapter(max_retries=retries),
        )
        self._request_session.headers.update(self.auth_headers)

    def _update_headers(self):
        self.auth_headers = {
            "Authorization": "Bearer " + get_token()
        }

    def _submit_task(
        self,
        task_type: APIEndpoints,
        payload: dict,
        request_id: str,
    ) -> Job:
        headers = self.auth_headers.copy()
        headers["X-Request-ID"] = request_id
        resp = self._request_session.post(
            urljoin(self.add_task_endpoint, task_type.value),
            timeout=self.request_timeout,
            headers=headers,
            json=payload,
            verify=not https_verification_disabled(),
        )
        if resp.ok:
            return Job(**resp.json())
        msg = f"[{resp.status_code}] Job not submitted: {resp.text}"
        raise ServerException(msg)

    def _get_job_status(self, request_id: str) -> Job:
        headers = self.auth_headers.copy()
        headers["X-Request-ID"] = request_id
        resp = self._request_session.get(
            self.status_endpoint,
            timeout=self.request_timeout,
            headers=headers,
            params={"request_id": request_id},
            verify=not https_verification_disabled(),
        )
        if resp.status_code == 404:
            msg = f"Request {request_id} not found."
            raise KeyError(msg)
        if not resp.ok:
            msg = f"[{resp.status_code}] {resp.text}"
            raise ServerException(msg)
        return Job(**resp.json())

    @override
    def set_token(self, token: str, validate: bool = True):
        """Persist the token, set it to env var and a file.
        Validate it.

        Args:
            token (str): auth token.
            validate (bool): you may switch off validation explicitly.
        """
        try:
            old_token = get_token()
        except RuntimeError as err:  # pragma: nocover
            Logging.warn(f"No token was set previously. {err.args[0]}", "set_token")
            old_token = ""  # nosec B105

        set_token(token)
        self._update_headers()
        self._request_session.headers.update(self.auth_headers)
        if validate:
            try:
                # try to obtain status of some job
                self._get_job_status("")
            except KeyError:
                # request is validated, even the key is missing
                pass
            except ServerException as exc:
                Logging.error(exc, "set_token")
                set_token(old_token)
                msg = f"Token failed validation: {exc.message}."
                raise ServerException(msg) from exc
            except requests.exceptions.ConnectionError as exc_coonection:
                message = "Could not validate token: cannot reach client."
                set_token(old_token)
                Logging.error(exc_coonection, "set_token")
                raise ServerException(message) from exc_coonection
        msg = (
            "Token successfully validated and "
            "will be used automatically in future sessions."
        )
        Logging.info(msg, "set_token")

    @override
    def execute(
        self,
        query_name: APIEndpoints,
        variable_values: dict,
        request_id: str,
    ) -> dict[str, Any]:
        job = self._submit_task(query_name, variable_values, request_id)
        try:
            # A client request ID is going to be different from
            # a server-generated one.
            Logging.info(f"Server created the job {job.request_id}", request_id)
            while job.status in [JobStatus.SUBMITTED.value, JobStatus.IN_PROGRESS.value]:
                time.sleep(APIv2Client.STATUS_CHECK_DELAY)
                Logging.info(
                    f"Job ({job.type}, {job.request_id}), status = {job.status}",
                    request_id
                )
                job = self._get_job_status(job.request_id)
            job.raise_on_error()
            Logging.info(
                f"Job ({job.type}, {job.request_id}) completed, status = {job.status}",
                request_id
            )
            return job.result
        except KeyboardInterrupt:
            count = self.kill(job.request_id)
            msg = f"Cancelled job {job.request_id} ({count} worker(s))."
            raise InterruptedError(msg)

    @override
    def kill(self, request_id: str) -> int:
        headers = self.auth_headers.copy()
        headers["X-Request-ID"] = request_id
        resp = self._request_session.delete(
            self.delete_endpoint,
            params={"request_id": request_id},
            headers=headers,
        )
        if resp.text and resp.text.isnumeric():
            return int(resp.text)
        return 0

    @override
    def generate_circuit(self, experiment_definition: QECExperimentDefinition, request_id: str) -> str:
        result = self.execute(
            query_name=APIEndpoints.GENERATE_CIRCUIT,
            variable_values=experiment_definition.to_json(),
            request_id=request_id,
        )
        return DataString.from_data_string(str(result.get("circuit"))).to_string()

    @override
    def simulate_circuit(self, stim_circuit: str | stim.Circuit, shots: int, request_id: str) -> tuple[Measurements, LeakageFlags | None]:
        result = self.execute(
            query_name=APIEndpoints.SIMULATE_CIRCUIT,
            variable_values={
                "circuit": str(DataString(str(stim_circuit))),
                "shots": shots,
            },
            request_id=request_id,
        )
        mmts = Measurements(
            DataString.from_data_string(str(result.get("measurements"))),
            data_format=DataFormat.F01,
        )
        leakage = None
        if result.get("leakage") is not None:
            leakage = LeakageFlags(
                DataString.from_data_string(str(result.get("leakage"))),
                DataFormat.F01,
            )
        return mmts, leakage

    @override
    def add_noise(self, stim_circuit: str | stim.Circuit, noise_model: NoiseModel, request_id: str) -> str:
        if noise_model.ENDPOINT is None:
            msg = f"Noise addition for {type(noise_model)} is not implemented."
            raise NotImplementedError(msg)
        result = self.execute(
            query_name=noise_model.ENDPOINT,
            variable_values={
                "circuit": str(DataString(str(stim_circuit))),
                "noise_model": noise_model.__dict__,
            },
            request_id=request_id,
        )
        return DataString.from_data_string(str(result.get("circuit"))).to_string()

    @override
    def decode(
        self,
        detectors: DetectionEvents,
        observables: ObservableFlips,
        decoder: Decoder,
        noisy_stim_circuit: str | stim.Circuit,
        leakage_flags: LeakageFlags | None,
        request_id: str,
    ) -> DecodingResult:
        query_name = APIEndpoints.DECODE if leakage_flags is None else APIEndpoints.DECODE_LEAKAGE
        job_result = self.execute(
            query_name=query_name,
            variable_values={
                "circuit": str(DataString(str(noisy_stim_circuit))),
                "detectors": detectors.as_data_string(DataFormat.B8),
                "observables": observables.as_data_string(DataFormat.B8),
                "leakage_flags": leakage_flags.as_data_string(DataFormat.B8) if leakage_flags else None,
                "decoder_type": decoder.decoder_type.value,
                "decoder_parameters": decoder.parameters,
                "decoder_jobs": decoder.parallel_jobs,
                "decoder_expgraph": decoder.use_experimental_graph,
            },
            request_id=request_id,
        )
        predictions = job_result.pop("predictions")
        job_result.pop("indices")
        result = DecodingResult(
            predictions_format=DataFormat.F01,
            **job_result,
        )
        result.predictions = BinaryDataType(predictions)
        return result

    @override
    def defect_rates(self,
        detectors: DetectionEvents,
        stim_circuit: str | stim.Circuit,
        request_id: str,
    ) -> dict[tuple[float, ...], list[float]]:
        query_name = APIEndpoints.DEFECT_RATES
        result = self.execute(
            query_name=query_name,
            variable_values={
                "circuit": str(DataString(str(stim_circuit))),
                "detectors": detectors.as_data_string(DataFormat.B8),
            },
            request_id=request_id,
        )
        rates = result["defect_rates"]
        return {
            tuple(map(float, pair["key"])): pair["value"]
            for pair in rates
        }

    @override
    def get_correlation_matrix_for_trimmed_data(
        self,
        detectors: DetectionEvents,
        noise_floor_circuit: str | stim.Circuit,
        use_default_noise_model_edges: bool,
        request_id: str,
    ) -> tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
        query_name = APIEndpoints.CORRELATION_MATRIX
        result = self.execute(
            query_name=query_name,
            variable_values={
                "circuit": str(DataString(str(noise_floor_circuit))),
                "detectors": detectors.as_data_string(DataFormat.B8),
                "use_stim_graph": use_default_noise_model_edges,
            },
            request_id=request_id,
        )
        matrix = result.get("correlation_matrix")
        qubit_to_detectors = cast(list, result.get("mapping", []))
        mapping = QubitCoordinateToDetectorMapping({
            tuple(qubit_det["qubit"]): qubit_det["detectors"] for qubit_det in qubit_to_detectors
        })
        return np.array(matrix, dtype=np.float64), mapping

    @override
    def trim_circuit_and_detectors(
        self,
        stim_circuit: str | stim.Circuit,
        detectors: DetectionEvents,
        request_id: str
    ) -> tuple[str, DetectionEvents]:
        query_name = APIEndpoints.TRIM_CIRCUIT_AND_DETECTORS
        result = self.execute(
            query_name=query_name,
            variable_values={
                "circuit": str(DataString(str(stim_circuit))),
                "detectors": detectors.as_data_string(DataFormat.B8),
            },
            request_id=request_id,
        )
        circuit = str(result.get("circuit"))
        dets_datastring = str(result.get("detectors"))
        detectors = DetectionEvents(DataString.from_data_string(dets_datastring), DataFormat.F01)
        return DataString.from_data_string(circuit).to_string(), detectors



# (c) Copyright Riverlane 2020-2025.
"""
Enum to encapsulate the different types of schedules available for
planar codes, and a function to return the X and Z schedules for a
planar code.
"""

from enum import Enum

from deltakit_circuit._basic_types import Coord2DDelta

from deltakit_explorer.codes._schedules._rotated_planar_code_schedules import (
    RotatedPlanarCodeSchedules,
)
from deltakit_explorer.codes._schedules._unrotated_planar_code_schedules import (
    UnrotatedPlanarCodeSchedules,
)

ScheduleClass = type[RotatedPlanarCodeSchedules] | type[UnrotatedPlanarCodeSchedules]


class ScheduleOrder(Enum):
    """
    Enum for specifying the schedule type for measuring stabilisers for a planar
    code.
    There are four options available:
        - ScheduleOrder.STANDARD, using, in general, the default_N_Z_dict
          schedules
        - ScheduleOrder.HORIZONTALLY_REFLECTED using, in general, the
          horizontally_reflected_N_Z_dict schedules
        - ScheduleOrder.VERTICALLY_REFLECTED using, in general, the
          vertically_reflected_N_Z_dict schedules
        - ScheduleOrder.DOUBLY_REFLECTED using, in general, the
          doubly_reflected_N_Z_dict schedules
    """

    STANDARD = "standard"
    HORIZONTALLY_REFLECTED = "horizontally_reflected"
    VERTICALLY_REFLECTED = "vertically_reflected"
    DOUBLY_REFLECTED = "doubly_reflected"


def get_x_and_z_schedules(
    schedule_class: ScheduleClass,
    schedule_order: ScheduleOrder,
    x_type_has_N_shape: bool = True,
) -> tuple[
    tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta],
    tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta],
]:
    """
    Function to retrieve the relevant schedule dictionary and assign the correct
    schedule to the X or Z plaquette. Returns a tuple containing first the X and
    then the Z schedules.

    Params
    ------
    schedule_class : ScheduleClass
        The class containing the schedules for the relevant code.
    schedule_order : ScheduleOrder
        The schedule order for the code.
    x_type_has_N_shape : bool, optional
        Bool specifying whether the X-type plaquette follows the N-type schedule,
        or if it should follow the Z-type schedule. The Z-type plaquette takes the
        other. By default, True.
    """
    if schedule_order == ScheduleOrder.STANDARD:
        dict_to_use = schedule_class.default_N_Z_dict
    elif schedule_order == ScheduleOrder.HORIZONTALLY_REFLECTED:
        dict_to_use = schedule_class.horizontally_reflected_N_Z_dict
    elif schedule_order == ScheduleOrder.VERTICALLY_REFLECTED:
        dict_to_use = schedule_class.vertically_reflected_N_Z_dict
    elif schedule_order == ScheduleOrder.DOUBLY_REFLECTED:
        dict_to_use = schedule_class.doubly_reflected_N_Z_dict
    else:
        msg = f"Did not recognise ScheduleOrder {schedule_order}"
        raise ValueError(msg)

    if x_type_has_N_shape:
        return dict_to_use["N"], dict_to_use["Z"]
    return dict_to_use["Z"], dict_to_use["N"]



# (c) Copyright Riverlane 2020-2025.
"""
This module includes implementation on a toy noise model.
Toy noise is a default choice of noise parameters to explore codes
scalability and to compare against other works. This a de-facto
experimental standard.
"""

from dataclasses import dataclass

from deltakit_circuit import measurement_noise_profile
from deltakit_circuit.gates import (
    TWO_QUBIT_GATES,
    OneQubitCliffordGate,
    OneQubitMeasurementGate,
    OneQubitResetGate,
)
from deltakit_circuit.noise_channels import Depolarise1, Depolarise2

from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters


@dataclass(kw_only=True)
class ToyNoise(NoiseParameters):
    """
    A noise model specified only by two parameters.

    One parameter gives the two-qubit operation error rate, which is assumed to be the
    dominant error. Therefore, all other noise channels (excluding the probability of
    obtaining an incorrect measurement result, which is specified separately) have an
    error rate a factor of ten lower. Idle noise is therefore not time dependent. All
    these noise channels are depolarising channels -- the noise channels occurring
    after a two-qubit gate are two-qubit depolarising channels and all others are
    one-qubit depolarising channels.

    The other parameter gives the probability of obtaining an incorrect measurement
    result.

    Parameters
    ----------
    p: float
        Depolarising error rate of two-qubit gates.
        One-qubit operations, including idle noise, have a depolarising error rate of
        0.1 * p. By default, this is 0.
    p_measurement_flip: Optional[float]
        The probability of obtaining an incorrect measurement result. By default, this
        has the same value as p.
    """

    name = "toy_noise"

    p: float = 0
    p_measurement_flip: float | None = None

    def __post_init__(self):
        self.p_measurement_flip = (
            self.p if self.p_measurement_flip is None else self.p_measurement_flip
        )

        depolarise1_generator = Depolarise1.generator_from_prob(self.p / 10)
        depolarise2_generator = Depolarise2.generator_from_prob(self.p)

        self.gate_noise.append(
            lambda noise_context: depolarise1_generator(
                noise_context.gate_layer_qubits(OneQubitCliffordGate)
            )
        )

        self.gate_noise.append(
            lambda noise_context: depolarise2_generator(
                noise_context.gate_layer_qubits(tuple(TWO_QUBIT_GATES))
            )
        )

        self.idle_noise = lambda qubit, _t=0.0: Depolarise1(
            qubit=qubit, probability=self.p / 10
        )

        # Measurement reset gates are not included here
        self.reset_noise.append(
            lambda noise_context: depolarise1_generator(
                noise_context.gate_layer_qubits(OneQubitResetGate)
            )
        )

        # Measurement reset gates are included here
        self.measurement_noise_after.append(
            lambda noise_context: depolarise1_generator(
                noise_context.gate_layer_qubits(OneQubitMeasurementGate)
            )
        )

        self.measurement_flip = measurement_noise_profile(self.p_measurement_flip)

    def __str__(self) -> str:
        if self.p_measurement_flip != self.p:
            return f"{self.name}_{self.p:.0e}_{self.p_measurement_flip:.0e}"
        return f"{self.name}_{self.p:.0e}"



# (c) Copyright Riverlane 2020-2025.
"""
This module stores an implementation of the unrotated planar code.
"""

import itertools

from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_types import Coord2D, Coord2DDelta

from deltakit_explorer.codes._planar_code._planar_code import PlanarCode, ScheduleType
from deltakit_explorer.codes._schedules import (
    ScheduleOrder,
    UnrotatedPlanarCodeSchedules,
    get_x_and_z_schedules,
)


class UnrotatedPlanarCode(PlanarCode):
    """
    Class representing the unrotated planar code. By default, Z logical is vertical
    (i.e. Z boundaries are horizontal) and X logical is horizontal (i.e. X boundaries
    are vertical). See the graph below for the default 3x3 unrotated code.

    This class also contains methods that help set up circuits for memory
    experiments.

    .. code-block:: text

        4├    ○ ---- Z ---- ○ ---- Z ---- ○
         │    |      |      |      |      |
        3├    X ---- ○ ---- X ---- ○ ---- X
         │    |      |      |      |      |
        2├    ○ ---- Z ---- ○ ---- Z ---- ○
         │    |      |      |      |      |
        1├    X ---- ○ ---- X ---- ○ ---- X
         │    |      |      |      |      |
        0├    ○ ---- Z ---- ○ ---- Z ---- ○
         │
         └----┴------┴------┴------┴------┴
              0      1      2      3      4

    Parameters
    ----------
    width : int
        The maximum number of data qubits in a single row.
    height : int
        The maximum number of data qubits in a single column.
    schedule_type : ScheduleType, optional
        The type of syndrome extraction schedule to use. By default,
        ScheduleType.SIMULTANEOUS.
    schedule_order : ScheduleOrder, optional
        The order of the syndrome extraction rounds, i.e, in which order to perform
        the consecutive entangling gates to measure the stabilier of a plaquette.
        By default, ScheduleOrder.STANDARD.
    use_ancilla_qubits : bool, optional
        Whether or not to use ancilla qubits for the stabilisers. By default, True.
    horizontal_stabiliser_with_top_left_is_z : bool, optional
        Whether the leftmost stabiliser in the top row is a Z stabiliser. By default,
        True.
    shift : Coord2DDelta, optional
        A vector by which to shift all of the qubit coordinates, by adding it to each
        qubit coordinate.
        By default, the 0 vector: Coord2DDelta(0, 0).
    """

    def __init__(
        self,
        width: int,
        height: int,
        schedule_type: ScheduleType = ScheduleType.SIMULTANEOUS,
        schedule_order: ScheduleOrder = ScheduleOrder.STANDARD,
        use_ancilla_qubits: bool = True,
        horizontal_stabiliser_with_top_left_is_z: bool = True,
        shift: Coord2DDelta = Coord2DDelta(0, 0),
    ):
        # coordinates of the bottom left vertex of the rectangle
        (self._x0, self._y0) = (0, 0)
        x_schedule, z_schedule = get_x_and_z_schedules(UnrotatedPlanarCodeSchedules, schedule_order)

        self._perform_css_checks = False

        self._horizontal_stabiliser_with_top_left_is_z = (
            horizontal_stabiliser_with_top_left_is_z
        )
        if self._horizontal_stabiliser_with_top_left_is_z:
            self.x_distance = width
            self.z_distance = height
        else:
            self.x_distance = height
            self.z_distance = width

        super().__init__(
            width=width,
            height=height,
            untransformed_x_schedule=x_schedule,
            untransformed_z_schedule=z_schedule,
            schedule_type=schedule_type,
            use_ancilla_qubits=use_ancilla_qubits,
            shift=shift,
        )

    def _calculate_untransformed_all_qubits(
        self,
    ) -> tuple[set[Qubit], set[Qubit], set[Qubit]]:
        # horizontal and vertical total length of the rectangle in which the data and ancillary qubits are.
        x_length = 2 * self.width - 1
        y_length = 2 * self.height - 1

        all_qubits = {
            Qubit(
                Coord2D(
                    x,
                    y,
                )
            )
            for x, y in itertools.product(
                range(self._x0, self._x0 + x_length),
                range(self._y0, self._y0 + y_length),
            )
        }

        # calculate all at the same time, so we only loop over all_qubits once
        x_ancilla, z_ancilla, data_qubits = set(), set(), set()

        # Calculate the parity of the x-coordinate of the left-most ancilla in the top row. This is of Z type in the default horizontal_stabiliser_with_top_left_is_z == True case, hence a reference.
        top_left_anc_x_mod_2 = (self._x0 + 1) % 2

        for qubit in all_qubits:
            x_coord_mod_2 = qubit.unique_identifier.x % 2
            y_coord_mod_2 = qubit.unique_identifier.y % 2
            # Data qubits are at coordinates of same parity
            if x_coord_mod_2 == y_coord_mod_2:
                data_qubits.add(qubit)
            # The rest of the qubits are all ancillas. If top row stabilisers are of Z type, then Z-ancillas's x-coord matches the reference, otherwise it doesn't.
            elif (
                self._horizontal_stabiliser_with_top_left_is_z
                and (x_coord_mod_2 == top_left_anc_x_mod_2)
            ) or (
                (not self._horizontal_stabiliser_with_top_left_is_z)
                and (x_coord_mod_2 != top_left_anc_x_mod_2)
            ):
                z_ancilla.add(qubit)
            else:
                x_ancilla.add(qubit)

        return x_ancilla, z_ancilla, data_qubits

    def _calculate_untransformed_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]:
        horizontal_logical_qubits = {
            Qubit(
                Coord2D(
                    x,
                    2 * self.height - 2,
                )
            )
            for x in range(0, 2 * self.width - 1, 2)
        }
        vertical_logical_qubits = {
            Qubit(Coord2D(0, y)) for y in range(0, 2 * self.height - 1, 2)
        }
        if self._horizontal_stabiliser_with_top_left_is_z:
            x_logical = {PauliX(qubit) for qubit in horizontal_logical_qubits}
            z_logical = {PauliZ(qubit) for qubit in vertical_logical_qubits}
        else:
            x_logical = {PauliX(qubit) for qubit in vertical_logical_qubits}
            z_logical = {PauliZ(qubit) for qubit in horizontal_logical_qubits}

        return (x_logical,), (z_logical,)

    def draw_patch(self, filename: str | None = None, unrotated_code: bool = True) -> None:
        return super().draw_patch(filename, unrotated_code)



# (c) Copyright Riverlane 2020-2025.
"""
Contains a few utility functions to be used in deltakit api
"""
from __future__ import annotations

import csv
import os
import sys
from pathlib import Path

APP_NAME = "deltakit-explorer"
DELTAKIT_SERVER_URL_ENV = "DELTAKIT_SERVER"
DELTAKIT_SERVER_DEFAULT_URL_ENV = "https://deltakit.riverlane.com/proxy"
HTTP_PACKET_LIMIT = 20_000_000


def get_log_directory() -> Path:
    """
    Get or generate the log directory path for the user.

    Returns:
        Path: The path to the log directory based on the order of availability.
    """

    # Check for 'LOG_DIRECTORY' environment variable
    env_folder = os.getenv('LOG_DIRECTORY')
    if env_folder:
        env_data_folder = Path(env_folder)
        if env_data_folder.exists() and env_data_folder.is_dir():
            return Path(env_folder)

    return Path.cwd()


def get_config_directory() -> Path:
    """Try to obtain OS-specific configuration folder for an app.
    Fallback scenario is to use current working directory."""
    try:
        config_dir_path, override_path = None, None
        if sys.platform == "win32":
            config_dir_path = Path.home() / "AppData" / "Local" / APP_NAME
            override_path = os.getenv("APPDATA")
        elif sys.platform == "darwin":
            config_dir_path = Path.home() / "Library" / "Application Support" / APP_NAME
        else:  # Linux and other Unix-like
            config_dir_path = Path.home() / ".config" / APP_NAME
            override_path = os.getenv("XDG_CONFIG_HOME")

        if override_path is not None:
            config_dir_path = Path(override_path) / APP_NAME
        if not config_dir_path.exists():
            config_dir_path.mkdir(parents=True)
        return config_dir_path
    except PermissionError:
        return Path.cwd()


def get_config_file_path() -> Path:
    """Path to a config file. Ensure existence of the folder."""
    return get_config_directory() / ".env"


def read_persisted_variables(file: Path) -> dict[str, str]:
    """
    Given a filename, returns content as a dictionary.
    Separator of key and value is `=`.
    Separator of items is a new line.

    Args:
        file (Path): path to a file with variables.

    Returns:
        Dict[str, str]:
            dictionary of persisted variables.
    """
    result = {}
    if file.exists():
        with Path.open(file, encoding="utf-8") as csv_stream:
            for row in csv.reader(csv_stream, delimiter="="):
                # empty line is ok
                if len(row) == 0:
                    continue
                # parsing of env file failed
                if len(row) != 2:
                    return {}
                result[row[0].strip()] = row[1].strip()
    return result


def set_variables(variables: dict[str, str], override: bool = False):
    """
    Set environment variables. If variable is defined,
    do not rewrite.

    Args:
        variables (Dict[str, str]): new values for environment variables.
        override (bool): override the variable.
    """
    for key, value in variables.items():
        if key not in os.environ or override:
            os.environ[key] = value


def load_environment_variables_from_drive():
    """
    If any environment variables are persisted, load them
    without overriding passed ones.
    """
    file = get_config_file_path()
    variables = read_persisted_variables(file)
    set_variables(variables, False)


def override_persisted_variables(variables: dict[str, str], file: Path):
    """
    Sets variables to the file. Overrides existing file.

    Args:
        variables (Dict[str, str]): new values to save in the file.
    """
    with Path.open(file, "w", encoding="utf-8") as csv_stream:
        writer = csv.writer(csv_stream, delimiter="=")
        for key, value in variables.items():
            writer.writerow([key, value])


def merge_variables(variables: dict[str, str], file: Path):
    """
    Adds or overrides the values in the file. Keeps old values.

    Args:
        variables (Dict[str, str]): new values to save in the file.
    """
    old_variables = read_persisted_variables(file)
    old_variables.update(variables)
    override_persisted_variables(old_variables, file)



# (c) Copyright Riverlane 2020-2025.
"""
This module defines a class for a CSS code. The class includes
multiple validity checks and provides methods to generate encoding,
stabiliser measurements, and logical measurement stages.
"""

from __future__ import annotations

import itertools
from collections import Counter
from collections.abc import Collection, Iterable, Sequence
from functools import cached_property

import galois
import numpy as np
from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._qubit_identifiers import PauliGate
from deltakit_circuit.gates import MX, MZ, RX, RZ
from numpy.typing import NDArray

from deltakit_explorer.codes._css._css_stage import CSSStage
from deltakit_explorer.codes._css._stabiliser_code import StabiliserCode
from deltakit_explorer.codes._logicals import (
    get_logical_operators_from_css_parity_check_matrices,
    get_logical_operators_from_tableau,
)
from deltakit_explorer.codes._stabiliser import Stabiliser


class CSSCode(StabiliserCode):
    """
    Class representing a CSS code.

    Parameters
    ----------
    stabilisers : Sequence[Sequence[Stabiliser]]
        Stabilisers defining the code. The i-th element of stabilisers contains an
        Iterable of stabilisers measured during the i-th time step. For instance:

        - if we intend to measure all stabilisers simultaneously, then this
          parameter could be simply given as [{...all stabilisers...}],
        - if we intend to do a spaced syndrome extraction, then this
          parameter could be simply given as
          [{...all x stabilisers...}, {...all z stabilisers...}].

        If any stabilisers do not have ancilla qubits, the `use_ancilla_qubits`
        parameter must be set to False.
    x_logical_operators : Optional[Sequence[Collection[PauliGate]]], optional
        An entry to provide manually the X logical operators of the CSS code. These
        operators must contain X Pauli terms only. If provided, then the Z logical
        operators must also be manually provided, otherwise a ValueError is raised.
        If `check_logical_operators_are_independent` is True, then each X logical operator
        must exactly only anticommute with the Z logical operator at the same index.
        There is also the possibility of automatically generating these logical operators,
        in which case neither should be provided and `calculate_logical_operators` should
        be set to True. To initialise with no logical operators, keep as None and set
        `calculate_logical_operators` to False.
        By default, None.
    z_logical_operators : Optional[Sequence[Collection[PauliGate]]], optional
        An entry to provide manually the Z logical operators of the CSS code. These
        operators must contain Z Pauli terms only. If provided, then the X logical
        operators must also be manually provided, otherwise a ValueError is raised.
        If `check_logical_operators_are_independent` is True, then each Z logical operator
        must exactly only anticommute with the X logical operator at the same index.
        There is also the possibility of automatically generating these logical operators,
        in which case neither should be provided and `calculate_logical_operators` should
        be set to True. To initialise with no logical operators, keep as None and set
        `calculate_logical_operators` to False.
        By default, None.
    use_ancilla_qubits : bool, optional
        Specifies whether ancilla qubits will be used for syndrome extraction. If
        this is True, the stabilisers must all have ancilla qubits and, in this case,
        the order the Paulis are given in the stabilisers will be the order in which
        information is extracted from the data qubits to the ancilla when performing
        syndrome extraction.
        By default, True.
    calculate_logical_operators: bool, optional,
        Whether to automatically calculate the logical operators. If this is True,
        then x_logical_operators and z_logical_operators should be None. Otherwise,
        a ValueError is raised.
        If `check_logical_operators_are_independent` is set to True, then the calculated
        logical operators will either be guaranteed to be independent or an error will
        be raised indicating that it didn't manage to meet this condition.
        By default, False.
    check_logical_operators_are_independent : bool, optional
        Whether to check if logical X-Z operator pairs are independent. If True, then
        the commuting relations [X_k, Z_l] = 0 (k != l) are checked for the logical
        operators, ensuring that the logical operators correspond to separate logical
        qubits. If False, this check is ignored, in which case the logical operators
        can be dependent, e.g. instead of only accepting X_1, X_2 and Z_1, Z_2, the
        input X_1, X_1*X_2 and Z_1, Z_1*Z_2 is also accepted.
        If both this flag and `calculate_logical_operators` are True, then the
        calculated logical operators will either be guaranteed to be independent or
        an error will be raised indicating that it didn't manage to meet this
        condition.
        By default, False.
    """

    def __init__(
        self,
        stabilisers: Sequence[Iterable[Stabiliser]],
        x_logical_operators: Sequence[Collection[PauliGate]] | None = None,
        z_logical_operators: Sequence[Collection[PauliGate]] | None = None,
        use_ancilla_qubits: bool = True,
        calculate_logical_operators: bool = False,
        check_logical_operators_are_independent: bool = False,
    ) -> None:
        super().__init__(
            stabilisers=stabilisers,
            x_logical_operators=x_logical_operators,
            z_logical_operators=z_logical_operators,
            use_ancilla_qubits=use_ancilla_qubits,
            check_logical_operators_are_independent=check_logical_operators_are_independent,
        )

        self._check_logical_operators_are_independent = (
            check_logical_operators_are_independent
        )
        self._data_qubit_index = {
            qubit: i
            for i, qubit in enumerate(
                sorted(self._data_qubits, key=lambda q: q.unique_identifier)
            )
        }

        self._x_stabilisers, self._z_stabilisers = self._calculate_x_and_z_stabilisers(
            self._stabilisers
        )

        if calculate_logical_operators:
            if (x_logical_operators is not None) or (z_logical_operators is not None):
                msg = "No logicals should be provided if calculate_logical_operators is set to True."
                raise ValueError(msg)
            (
                x_logical_operators,
                z_logical_operators,
            ) = self._calculate_logical_operators()
            self.overwrite_logicals(x_logical_operators, z_logical_operators)

        if getattr(self, "_perform_css_checks", True):
            self._check_duplicate_stabilisers(self._stabilisers)
            self._check_for_duplicate_paulis(
                x_logical_operators if x_logical_operators is not None else [],
                z_logical_operators if z_logical_operators is not None else [],
            )
            self._check_stabiliser_and_logical_operator_types(
                self._stabilisers, self.x_logical_operators, self.z_logical_operators
            )
            self._check_logical_operators_are_supported_on_data_qubits(
                self.x_logical_operators + self.z_logical_operators, self._data_qubits
            )
            self._check_ancilla_qubit_properties(
                self._ancilla_qubits, self._data_qubits, self._stabilisers
            )
            self._check_commutation_relations(
                self._x_stabilisers,
                self._z_stabilisers,
                self.x_logical_operators,
                self.z_logical_operators,
                self._check_logical_operators_are_independent,
            )
            if self._use_ancilla_qubits:
                self._check_stabiliser_lengths(self._stabilisers)
                self._check_unique_data_qubits_in_layers(self._stabilisers)
                self._check_schedule_is_valid(self._x_stabilisers, self._z_stabilisers)

    @staticmethod
    def x_and_z_operators_commute(
        x_operator: Iterable[PauliX | None], z_operator: Iterable[PauliZ | None]
    ) -> bool:
        """
        Compute whether an X-type stabiliser commutes with a Z-type stabiliser.
        """
        anticommutation_counter = Counter(
            (paulix.qubit == pauliz.qubit)
            for paulix in x_operator
            if paulix is not None
            for pauliz in z_operator
            if pauliz is not None
        )

        return (anticommutation_counter[True] % 2) == 0

    @staticmethod
    def _check_duplicate_stabilisers(
        stabilisers: tuple[tuple[Stabiliser, ...], ...]
    ) -> None:
        for ind_lay, simultaneous_stabilisers in enumerate(stabilisers):
            if len(simultaneous_stabilisers) != len(set(simultaneous_stabilisers)):
                msg = f"Layer {ind_lay} of stabilisers contains duplicates."
                raise ValueError(msg)

    @staticmethod
    def _check_for_duplicate_paulis(
        x_logical_operators: Sequence[Collection[PauliGate]],
        z_logical_operators: Sequence[Collection[PauliGate]],
    ) -> None:
        """
        Check if there are any PauliGate duplicates in any of the logical operators.
        """
        for x_log_op in x_logical_operators:
            if len(set(x_log_op)) < len(x_log_op):
                msg = "One of the X-logical operators contains duplicate PauliX objects."
                raise ValueError(msg)
        for z_log_op in z_logical_operators:
            if len(set(z_log_op)) < len(z_log_op):
                msg = "One of the Z-logical operators contains duplicate PauliZ objects."
                raise ValueError(msg)

    @staticmethod
    def _check_stabiliser_and_logical_operator_types(
        stabilisers: tuple[tuple[Stabiliser, ...], ...],
        x_logical_operators: tuple[frozenset[PauliGate], ...],
        z_logical_operators: tuple[frozenset[PauliGate], ...],
    ) -> None:
        """
        Check the following:

            1. each stabiliser has either only X or only Z Pauli terms,

            2. the number of stabilisers is at least one,

            3. the lengths of x_logical_operators and z_logical_operators are equal,

            4. none of the logical operators is the identity,

            5. each X logical operator has only X Pauli terms, and

            6. each Z logical operator has only Z Pauli terms.
        """
        # Check condition 1)
        for simultaneous_stabilisers in stabilisers:
            for stabiliser in simultaneous_stabilisers:
                pauli_types = {
                    type(pauli) for pauli in stabiliser.paulis if pauli is not None
                }
                if len(pauli_types) != 1 or not pauli_types.issubset({PauliX, PauliZ}):
                    msg = (
                        "CSSCode object was initialised with incorrect type "
                        "stabilisers. Each stabiliser must consist of either all X or "
                        "all Z Paulis terms."
                    )
                    raise ValueError(msg)
        # Check condition 2)
        if (
            sum(
                len(simultaneous_stabilisers)
                for simultaneous_stabilisers in stabilisers
            )
            == 0
        ):
            msg = "CSSCode object was initialised with no stabilisers."
            raise ValueError(msg)
        # Check condition 3)
        if len(x_logical_operators) != len(z_logical_operators):
            msg = (
                "The lengths of x_logical_operators and z_logical_operators are not "
                "equal."
            )
            raise ValueError(msg)
        # Check condition 4)
        if any(len(x_logical) == 0 for x_logical in x_logical_operators):
            msg = (
                "x_logical_operators contains an empty Iterable. Identity cannot be a"
                " logical operator."
            )
            raise ValueError(msg)
        if any(len(z_logical) == 0 for z_logical in z_logical_operators):
            msg = (
                "z_logical_operators contains an empty Iterable. Identity cannot be a"
                " logical operator."
            )
            raise ValueError(msg)
        # Check condition 5)
        for x_logical in x_logical_operators:
            if not all(isinstance(pauli, PauliX) for pauli in x_logical):
                msg = "All X logical operators should consist of only X Pauli terms."
                raise ValueError(msg)
        # Check condition 6)
        for z_logical in z_logical_operators:
            if not all(isinstance(pauli, PauliZ) for pauli in z_logical):
                msg = "All Z logical operators should consist of only Z Pauli terms."
                raise ValueError(msg)

    @staticmethod
    def _calculate_x_and_z_stabilisers(
        stabilisers: tuple[tuple[Stabiliser, ...], ...],
    ) -> tuple[list[set[Stabiliser]], list[set[Stabiliser]]]:
        """
        Return two lists, one containing the X-stabilisers and the other the
        Z-stabilisers.

        Returns
        -------
        Tuple[List[Set[Stabiliser]], List[Set[Stabiliser]]]
            Tuple containing X-stabilisers and Z-stabilisers.
        """
        x_stabilisers: list[set[Stabiliser]] = []
        z_stabilisers: list[set[Stabiliser]] = []
        for simultaneous_stabilisers in stabilisers:
            x_layer = set()
            z_layer = set()
            for stabiliser in simultaneous_stabilisers:
                if isinstance(next(iter(stabiliser.operator_repr)), PauliX):
                    x_layer.add(stabiliser)
                else:
                    z_layer.add(stabiliser)
            x_stabilisers.append(x_layer)
            z_stabilisers.append(z_layer)
        return x_stabilisers, z_stabilisers

    @staticmethod
    def check_logical_operators_are_independent(
        x_logicals: Iterable[Iterable[PauliX]], z_logicals: Iterable[Iterable[PauliZ]]
    ):
        """
        Check that the logical operators are independent. In other words, check that
        each logical operator commutes with all other logical operators except for
        their pair (i.e. the logical operator of opposite type located at the same
        index). Raise a ValueError otherwise.
        """
        logicals_zipped = list(zip(x_logicals, z_logicals))
        for i, (x_logical, z_logical) in enumerate(logicals_zipped):
            if CSSCode.x_and_z_operators_commute(x_logical, z_logical):
                msg = (
                    f"The X and Z logical operators with the same index should anticommute,"
                    f" but at index {i} found {x_logical} and {z_logical}."
                )
                raise ValueError(msg)

            for other_x_logical, other_z_logical in logicals_zipped[i + 1 :]:
                for logical, other_logical in (
                    (x_logical, other_z_logical),
                    (z_logical, other_x_logical),
                ):
                    if not CSSCode.x_and_z_operators_commute(logical, other_logical):
                        msg = (
                            "Two logical operators with different indices should commute,"
                            f" but found operators {logical} and {other_logical}."
                        )
                        raise ValueError(msg)

    @staticmethod
    def _check_commutation_relations(
        x_stabilisers: list[set[Stabiliser]],
        z_stabilisers: list[set[Stabiliser]],
        x_logical_operators: tuple[frozenset[PauliGate], ...],
        z_logical_operators: tuple[frozenset[PauliGate], ...],
        check_logical_operators_are_independent: bool,
    ) -> None:
        """
        Check the commutation-anticommutation relations between stabilisers and logical
        operators. More precisely, check that:

            1. all stabilisers commute with each other,

            2. each logical operator commutes with each stabiliser, and

            3. each logical operator commutes with each logical operator except one
                (i.e. the other type of logical operator on the same logical qubit).
        """
        x_stabilisers_flat = [stab for sim_stabs in x_stabilisers for stab in sim_stabs]
        z_stabilisers_flat = [stab for sim_stabs in z_stabilisers for stab in sim_stabs]

        # Check condition 1)
        for x_stab, z_stab in itertools.product(x_stabilisers_flat, z_stabilisers_flat):
            if not CSSCode.x_and_z_operators_commute(x_stab.paulis, z_stab.paulis):
                msg = (
                    "CSSCode object was initialised with anticommuting "
                    "stabilisers. Namely, the X-stabiliser defined on data qubits "
                    f"{x_stab.data_qubits} anticommutes with the Z-stabiliser "
                    f"defined on data qubits {z_stab.data_qubits}."
                )
                raise ValueError(msg)
        # Check condition 2)
        for x_stab, z_log in itertools.product(x_stabilisers_flat, z_logical_operators):
            if not CSSCode.x_and_z_operators_commute(x_stab.paulis, z_log):
                msg = (
                    "CSSCode object was initialised with a Z-logical operator that "
                    "anticommutes with a stabiliser. Namely, the "
                    f"X-stabiliser defined on data qubits {x_stab.data_qubits} "
                    f"anticommutes with the Z-logical operator {z_log}."
                )
                raise ValueError(msg)
        for z_stab, x_log in itertools.product(z_stabilisers_flat, x_logical_operators):
            if not CSSCode.x_and_z_operators_commute(z_stab.paulis, x_log):
                msg = (
                    "CSSCode object was initialised with an X-logical operator that "
                    "anticommutes with a stabiliser. Namely, the "
                    f"Z-stabiliser defined on data qubits {z_stab.data_qubits} "
                    "anticommutes with the "
                    f"X-logical operator {x_log} "
                )
                raise ValueError(msg)
        # Check condition 3)
        for (x_ind, x_log), (z_ind, z_log) in itertools.product(
            enumerate(x_logical_operators), enumerate(z_logical_operators)
        ):
            if x_ind == z_ind and CSSCode.x_and_z_operators_commute(x_log, z_log):
                msg = (
                    "CSSCode object was initialised with commuting X- and Z-logical "
                    f"operators defined on the logical qubit at index {x_ind}. "
                    f"The X-logical operator at index {x_ind} "
                    f"is {x_log}, and the Z-logical operator at index {z_ind} is "
                    f"{z_log}."
                )
                raise ValueError(msg)
            if (
                check_logical_operators_are_independent
                and x_ind != z_ind
                and not CSSCode.x_and_z_operators_commute(x_log, z_log)
            ):
                msg = (
                    "CSSCode object was initialised with anticommuting X- and "
                    "Z-logical operators that act on different logical qubits at "
                    f"indices {x_ind, z_ind}. The X-logical operator at index {x_ind} "
                    f"is {x_log}, and the Z-logical operator at index {z_ind} is "
                    f"{z_log}."
                )
                raise ValueError(msg)

    @staticmethod
    def _check_ancilla_qubit_properties(
        ancilla_qubits: set[Qubit],
        data_qubits: set[Qubit],
        stabilisers: tuple[tuple[Stabiliser, ...], ...],
    ) -> None:
        """
        Check the following properties for ancilla qubits:

            1. data qubits are disjoint from ancilla qubits,

            2. each set of stabilisers measured simultaneously has no repeated
            ancilla qubits.
        """
        if len(ancilla_qubits) > 0:
            if not data_qubits.isdisjoint(ancilla_qubits):
                msg = (
                    f"The following ancilla qubits are also used as data qubits: "
                    f"{data_qubits.intersection(ancilla_qubits)}"
                )
                raise ValueError(msg)
            for ind, simultaneous_stabilisers in enumerate(stabilisers):
                simultaneous_ancillas = {
                    stab.ancilla_qubit for stab in simultaneous_stabilisers
                }
                if len(simultaneous_ancillas) != len(simultaneous_stabilisers):
                    msg = (
                        f"There are duplicate ancilla qubits in layer {ind} of "
                        "stabilisers."
                    )
                    raise ValueError(msg)

    @staticmethod
    def _check_logical_operators_are_supported_on_data_qubits(
        logical_operators: tuple[frozenset[PauliGate], ...], data_qubits: set[Qubit]
    ) -> None:
        """
        Check that the logical operators are defined on the code's data qubits.
        """
        log_ops_qubits = set()
        for log_op in logical_operators:
            for pauli in log_op:
                log_ops_qubits.add(pauli.qubit)
        if not log_ops_qubits.issubset(data_qubits):
            msg = (
                "Some logical operators are not supported on the CSS code's data "
                "qubits."
            )
            raise ValueError(msg)

    @staticmethod
    def _check_stabiliser_lengths(
        stabilisers: tuple[tuple[Stabiliser, ...], ...],
    ) -> None:
        """
        Assuming all stabilisers are measured using syndrome extraction circuits, check
        if each Set of the Sequence contains Stabilisers with paulis attribute of the
        same length. (It is, however, possible for different Sets to contain
        Stabilisers of different lengths.)
        """
        for ind, simultaneous_stabilisers in enumerate(stabilisers):
            # raise an error if the lengths of paulis in the current stabiliser layer
            # are not all the same
            if (
                len(
                    {
                        len(stabiliser.paulis)
                        for stabiliser in simultaneous_stabilisers
                    }
                )
                > 1
            ):
                msg = (
                    f"Layer {ind} of stabilisers contains two elements whose paulis "
                    "attributes are of different lengths."
                )
                raise ValueError(msg)

    @staticmethod
    def _check_unique_data_qubits_in_layers(
        stabilisers: tuple[tuple[Stabiliser, ...], ...],
    ):
        """
        Assuming all stabilisers are measured using syndrome extraction circuits, check
        that stabilisers is specified in such a way that during syndrome extraction
        none of the data qubits is used multiple times in the same layer.
        """
        for ind, simultaneous_stabilisers in enumerate(stabilisers):
            if len(simultaneous_stabilisers) == 0:
                continue
            current_paulis_len = len(next(iter(simultaneous_stabilisers)).paulis)
            for layer_ind in range(current_paulis_len):
                paulis_and_nones_in_layer = [
                    # navigate down "columns" of 2d array, where column=layer
                    stab.paulis[layer_ind]
                    for stab in simultaneous_stabilisers
                ]
                qubits_in_layer = [
                    pauli.qubit
                    for pauli in paulis_and_nones_in_layer
                    if pauli is not None
                ]
                if len(qubits_in_layer) != len(set(qubits_in_layer)):
                    msg = (
                        f"Layer {ind} of stabilisers contains at least two Stabiliser "
                        f"objects that have the same qubit at index {layer_ind} of "
                        "their paulis attributes. This means that the syndrome "
                        "extraction circuit is invalid, as each layer should "
                        "contain unique qubits only."
                    )
                    raise ValueError(msg)

    @staticmethod
    def _check_schedule_is_valid(
        x_stabilisers: list[set[Stabiliser]], z_stabilisers: list[set[Stabiliser]]
    ) -> None:
        """
        Assuming all stabilisers are measured using syndrome extraction circuits,
        check that the schedules define a valid syndrome extraction for the code. This
        assumes that _check_unique_data_qubits_in_layers already passed.
        """

        def _entangled_ancillas(x_stab: Stabiliser, z_stab: Stabiliser) -> bool:
            """
            Compute whether the schedule between an X- and Z-type stabiliser pair
            entangles the ancilla qubits or not, assuming that their paulis attributes
            have the same length. Return the bool for later reference.
            """
            x_ordered_qubits = [
                pauli.qubit if pauli is not None else None for pauli in x_stab.paulis
            ]
            z_ordered_qubits = [
                pauli.qubit if pauli is not None else None for pauli in z_stab.paulis
            ]
            # Check twist. Imagine we pull all CX gates before the CZ gates and each
            # time for a data qubit (target qubit) CX happens after CZ, we switch
            # is_twisted, as that means a CZ between the two ancillas has to be
            # applied.
            is_twisted = False
            for ind, qubit in enumerate(x_ordered_qubits):
                if qubit is None:
                    continue
                if qubit in z_ordered_qubits[:ind]:
                    is_twisted = not is_twisted
            return is_twisted

        twisted_info = []
        for x_simultaneous_stabilisers, z_simultaneous_stabilisers in zip(
            x_stabilisers, z_stabilisers
        ):
            twisted_info_layer = []
            for x_stab, z_stab in itertools.product(
                x_simultaneous_stabilisers, z_simultaneous_stabilisers
            ):
                if _entangled_ancillas(x_stab, z_stab):
                    twisted_info_layer.append(
                        [x_stab.ancilla_qubit, z_stab.ancilla_qubit]
                    )
            twisted_info.append(twisted_info_layer)

        if any(len(twisted_info_layer) > 0 for twisted_info_layer in twisted_info):
            # Build error string
            error_str = ""
            for layer_ind, twisted_info_layer in enumerate(twisted_info):
                if len(twisted_info_layer) > 0:
                    error_str += "\n" + " " * 4 + f"in layer {layer_ind} of stabilisers"
                for ancilla_pair in twisted_info_layer:
                    x_anc, z_anc = ancilla_pair
                    error_str += "\n" + " " * 8 + str(x_anc) + " and " + str(z_anc)

            raise ValueError(
                "The scheduling of the stabilisers doesn't give a valid syndrome "
                "extraction circuit. More precisely, the provided schedule entangles "
                "the following ancilla pairs:" + error_str
            )

    def encode_logical_zeroes(self) -> CSSStage:
        if len(self.z_logical_operators) == 0:
            msg = (
                "No logical Z operators provided so cannot prepare logical zeroes "
                "state."
            )
            raise NotImplementedError(msg)
        return CSSStage(final_round_resets=[RZ(qubit) for qubit in self._data_qubits])

    def encode_logical_pluses(self) -> CSSStage:
        if len(self.x_logical_operators) == 0:
            msg = (
                "No logical X operators provided so cannot prepare logical pluses "
                "state."
            )
            raise NotImplementedError(msg)
        return CSSStage(final_round_resets=[RX(qubit) for qubit in self._data_qubits])

    def measure_z_logicals(self) -> CSSStage:
        if len(self.z_logical_operators) == 0:
            msg = "No logical Z operators provided so cannot measure logical Z operators."
            raise NotImplementedError(msg)
        return CSSStage(
            first_round_measurements=[MZ(qubit) for qubit in self._data_qubits],
            observable_definitions={
                ind: [pauli.qubit for pauli in z_logical]
                for ind, z_logical in enumerate(self.z_logical_operators)
            },
        )

    def measure_x_logicals(self) -> CSSStage:
        if len(self.x_logical_operators) == 0:
            msg = "No logical X operators provided so cannot measure logical X operators."
            raise NotImplementedError(msg)
        return CSSStage(
            first_round_measurements=[MX(qubit) for qubit in self._data_qubits],
            observable_definitions={
                ind: [pauli.qubit for pauli in x_logical]
                for ind, x_logical in enumerate(self.x_logical_operators)
            },
        )

    @staticmethod
    def from_matrix(
        h_x: NDArray = np.zeros((0, 0)),
        h_z: NDArray = np.zeros((0, 0)),
        log_x_ops: NDArray = np.zeros((0, 0)),
        log_z_ops: NDArray = np.zeros((0, 0)),
    ) -> CSSCode:
        """
        Return a CSSCode object with stabilisers that have no ancilla qubits. The
        inputs must be 0-1 matrices.

        Parameters
        ----------
        h_x : NDArray, optional
            The check matrix (containing only 0 and 1) for X stabilisers where each
            row represents an X stabiliser. If an empty matrix, then this means the
            CSS code has no X stabilisers. By default, np.zeros((0,0)).
        h_z : NDArray, optional
            The check matrix (containing only 0 and 1) for Z stabilisers where each
            row represents a Z stabiliser. If an empty matrix, then this means the
            CSS code has no Z stabilisers. By default, np.zeros((0,0)).
        log_x_ops : NDArray, optional
            Matrix (containing only 0 and 1) representing the logical X operators
            where each row represents a logical X operator. If an empty matrix,
            then this means the CSS code was specified without X logical
            operators. By default, np.zeros((0,0)).
        log_z_ops : NDArray, optional
            Matrix (containing only 0 and 1) representing the logical Z operators
            where each row represents a logical Z operator. If an empty matrix,
            then this means the CSS code was specified without Z logical
            operators. By default, np.zeros((0,0)).

        Returns
        -------
        CSSCode
            The CSS code obtained from the matrix input.

        Raises
        ------
        ValueError
            If both matrices h_x and h_z are empty.
        ValueError
            If neither of the matrices h_x_and h_z is empty but they have different
            numbers of columns.
        ValueError
            If the numbers of rows differ for the matrices log_x_ops and log_z_ops.
        ValueError
            If either log_x_ops or log_z_ops has a different number of columns to
            h_x or h_z.
        ValueError
            If any of the inputs have elements that are different from 0 and 1.
        """
        num_x_stabs, num_qubits_x = np.shape(h_x)
        num_z_stabs, num_qubits_z = np.shape(h_z)

        # Size checks for h_x and h_z
        if num_qubits_x == 0 and num_qubits_z == 0:
            msg = "The matrices h_x and h_z cannot both be empty."
            raise ValueError(msg)
        if num_qubits_x > 0 and num_qubits_z > 0 and num_qubits_x != num_qubits_z:
            msg = "The matrices h_x and h_z need to have the same number of columns."
            raise ValueError(msg)

        num_qubits = num_qubits_x if num_qubits_x > 0 else num_qubits_z
        num_x_log_ops, num_qubits_log_x = np.shape(log_x_ops)
        num_z_log_ops, num_qubits_log_z = np.shape(log_z_ops)

        # Size checks for log_x_ops and log_z_ops
        if num_x_log_ops != num_z_log_ops:
            msg = (
                "The matrices log_x_ops and log_z_ops cannot have different numbers of "
                "rows."
            )
            raise ValueError(msg)
        if num_qubits_log_x not in [0, num_qubits]:
            msg = "The matrix log_x_ops has a different number of columns to h_x or h_z."
            raise ValueError(msg)
        if num_qubits_log_z not in [0, num_qubits]:
            msg = "The matrix log_z_ops has a different number of columns to h_x or h_z."
            raise ValueError(msg)

        def _check_and_return_entry(entry):
            if entry not in [0, 1]:
                msg = "Some elements in the provided matrices are not 0 or 1."
                raise ValueError(msg)
            return entry

        x_stabilisers = (
            []
            if num_x_stabs == 0 or num_qubits_x == 0
            else [
                Stabiliser(
                    paulis=[
                        PauliX(j)
                        for j in range(num_qubits)
                        if _check_and_return_entry(h_x[i, j]) == 1
                    ]
                )
                for i in range(num_x_stabs)
            ]
        )
        z_stabilisers = (
            []
            if num_z_stabs == 0 or num_qubits_z == 0
            else [
                Stabiliser(
                    paulis=[
                        PauliZ(j)
                        for j in range(num_qubits)
                        if _check_and_return_entry(h_z[i, j]) == 1
                    ]
                )
                for i in range(num_z_stabs)
            ]
        )
        x_logical_operators = (
            []
            if num_x_log_ops == 0 or num_qubits_log_x == 0
            else [
                [
                    PauliX(j)
                    for j in range(num_qubits)
                    if _check_and_return_entry(log_x_ops[i, j]) == 1
                ]
                for i in range(num_x_log_ops)
            ]
        )
        z_logical_operators = (
            []
            if num_z_log_ops == 0 or num_qubits_log_z == 0
            else [
                [
                    PauliZ(j)
                    for j in range(num_qubits)
                    if _check_and_return_entry(log_z_ops[i, j]) == 1
                ]
                for i in range(num_z_log_ops)
            ]
        )
        return CSSCode(
            stabilisers=[x_stabilisers + z_stabilisers],
            x_logical_operators=x_logical_operators,
            z_logical_operators=z_logical_operators,
            use_ancilla_qubits=False,
        )

    @cached_property
    def parity_check_matrices(self) -> tuple[NDArray, NDArray]:
        """
        Construct parity check matrices Hx and Hz from the CSS Code.

        Note that neither logical operators nor ancilla qubits are contained in
        the parity check matrices.

        Returns
        -------
        Tuple[NDArray, NDArray]
            A tuple of two matrices, Hx and Hz.
        """
        # first, assign an index to each data qubit which will be the index
        # of the column in the matrix. qubits may have strange coordinates or be
        # transformed, so we cannot rely on the qubit coordinates
        all_x_stabilisers = list(itertools.chain.from_iterable(self._x_stabilisers))
        all_z_stabilisers = list(itertools.chain.from_iterable(self._z_stabilisers))
        x_parity_mat, z_parity_mat = np.zeros(
            (len(all_x_stabilisers), len(self.data_qubits)), dtype=np.uint8
        ), np.zeros((len(all_z_stabilisers), len(self.data_qubits)), dtype=np.uint8)

        # turn stabilisers into rows of the matrices.
        # note that since stabilisers are kept in Sets, the order
        # in which we iterate over them is non-deterministic, and so
        # is the order of the rows of the parity check matrices
        for matrix, stab_type in zip(
            (x_parity_mat, z_parity_mat), (all_x_stabilisers, all_z_stabilisers)
        ):
            for stab_index, stab in enumerate(stab_type):
                for pauli in (p for p in stab.paulis if p is not None):
                    matrix[stab_index, self._data_qubit_index[pauli.qubit]] = 1

        return x_parity_mat, z_parity_mat

    def _calculate_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]:
        """
        Compute the logical operators of the code. First tries doing so by completing
        a stim Tableau, and if it doesn't yield X logicals purely made of X terms
        and Z logicals purely made of Z terms, falls back to doing so via the parity
        check matrix, for which that condition is guaranteed (but not necessarily
        independence).

        Tries to return independent logical operators but doesn't check whether they
        are.

        Returns
        -------
        tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]
            The logical operators, provided as a tuple of all the X logical
            operators at index 0 and all the Z logical operators at index 1.
        """
        # First try computing from Tableau (guarantees independent logical operators).
        x_logicals, z_logicals = get_logical_operators_from_tableau(
            stabilisers=[
                stabiliser
                for simultaneous_stabilisers in self._stabilisers
                for stabiliser in simultaneous_stabilisers
            ],
            num_logical_qubits=self.calculate_number_of_logical_qubits(),
        )

        # Check X logical operators are made only of X gates and Z logical operators are
        # made only of Z gates (assumed by CSSCode).
        x_logicals_are_x_only = all(
            isinstance(pauli, PauliX) for logical in x_logicals for pauli in logical
        )
        z_logicals_are_z_only = all(
            isinstance(pauli, PauliZ) for logical in z_logicals for pauli in logical
        )

        if x_logicals_are_x_only and z_logicals_are_z_only:
            return x_logicals, z_logicals

        # Otherwise, fallback to method from BPOSD package, which guarantees X/Z logicals are X/Z-only
        # but not independence.
        x_logicals, z_logicals = get_logical_operators_from_css_parity_check_matrices(
            *self.parity_check_matrices,
            column_to_qubit={v: k for k, v in self._data_qubit_index.items()},
        )

        return x_logicals, z_logicals

    def calculate_number_of_logical_qubits(self) -> int:
        """
        Calculate the maximum number of logical qubits that are encoded into the code by
        subtracting the rank of Hx and Hz parity matrices from
        the number of (data) qubits.

        Note: this might not agree with the number of logicals that the user specified
        when initialising the `CSSCode` instance.
        """
        hx_mat, hz_mat = self.parity_check_matrices

        return len(self.data_qubits) - int(
            np.linalg.matrix_rank(galois.GF2(hx_mat))
            + np.linalg.matrix_rank(galois.GF2(hz_mat))
        )



# (c) Copyright Riverlane 2020-2025.
"""This file contains types to be used as input fields
for the generate_circuit mutation
"""

from __future__ import annotations

import dataclasses
import enum
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from io import BytesIO, StringIO, TextIOWrapper
from math import ceil
from pathlib import Path
from typing import Any, ClassVar

import numpy as np
import numpy.typing as npt
import stim

from deltakit_explorer.data._converter import (
    read_01,
    read_b8,
    read_csv,
    write_01,
    write_b8,
    write_csv,
)
from deltakit_explorer.enums._api_enums import APIEndpoints, DataFormat, DecoderType
from deltakit_explorer.types._data_string import DataString


class JSONable:
    """Provide a utility static method ``to_json`` to inheriting classes."""

    @staticmethod
    def _to_json(obj) -> Any:
        """Convert dataclass or object to dict for JSON serialisation."""
        if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
            return JSONable._to_json(dataclasses.asdict(obj))
        if isinstance(obj, dict):
            return {key: JSONable._to_json(value) for key, value in obj.items()}
        if isinstance(obj, list):
            return [JSONable._to_json(value) for value in obj]
        if isinstance(obj, enum.Enum):
            return JSONable._to_json(obj.value)
        if isinstance(obj, (str, int, float)):
            return obj
        if obj is None:
            return None
        msg = f"Object of type {type(obj)} is not a dataclass or serialisable object."
        raise ValueError(msg)

    def to_json(self) -> dict:
        return JSONable._to_json(self)


# pylint: disable=duplicate-code, too-few-public-methods, invalid-name
@dataclass
class MatrixSpecifications(JSONable):
    """Parameters to be passed to BivariateBicycle code type, IBM Bivariate
    Bicycle qLDPC codes, as specified in arXiv:2308.07915.
    The parameter `n` below as in the [[n,k,d]] specification of an
    error-correcting code refers to a number of physical qubits.

    Args:
        param_l (int):
            The parameter `l`, used to construct a code of length `n = 2lm`.
        param_m (int):
            The parameter `m`, used to construct a code of length `n = 2lm`.
        m_A_powers (List[int]):
            The powers of the polynomial terms used to construct
            the matrix A. Each polynomial is of the form
            `x^a + y^b + y^c` so the sequence should specify
            `[a, b, c]`.
        m_B_powers (List[int]):
            The powers of the polynomial terms used to construct
            the matrix B. Each polynomial is of the form
            `y^a + x^b + x^c` so the sequence should specify
            `[a, b, c]`.
    """

    param_l: int
    param_m: int
    m_A_powers: list[int]
    m_B_powers: list[int]

    def to_gql(self):
        """Get a dictionary for the current object"""

        return {
            "paramL": self.param_l,
            "paramM": self.param_m,
            "mAPowers": self.m_A_powers,
            "mBPowers": self.m_B_powers,
        }


# pylint: disable=duplicate-code, too-few-public-methods
@dataclass
class Sizes(JSONable):
    """The size parameter to be passed to generate_circuit() method.

    Args:
        sizes (Sequence[int]):
            - QUANTUM_MEMORY: `(x_distance, z_distance)` for RotatedPlanarCode,
              `(distance,)` for repetition, `(width, height)` for UnrotatedPlanarCode.
            - STABILITY: patch size `(width, height)`.

    """

    sizes: Sequence[int]

    def to_gql(self):
        """Get a dictionary for the current object"""

        return {"sizes": self.sizes}


# pylint: disable=duplicate-code, too-few-public-methods
@dataclass
class CircuitParameters(JSONable):
    """Parameter to be passed as a type for additional parameters
    for generate_circuit"""

    matrix_specifications: MatrixSpecifications | None = None
    sizes: Sizes | None = None

    @staticmethod
    def from_sizes(sizes: Iterable[int]) -> CircuitParameters:
        """Use with Repetition, Planar and Toric codes. Defines parameters
        for a circuit, which is initialised by a grid. Repetition code accepts
        a single number, other codes - two.

        Args:
            sizes (Iterable[int]): List or tuple of integer sizes.

        Return:
            CircuitParameters: parameters object for circuit generation.

        Examples:
            Creation and usage of parameters::

                circuit = client.generate_circuit(
                    types.QECExperimentDefinition(
                        experiment_type=types.QECExperimentType.QUANTUM_MEMORY,
                        code_type=types.QECECodeType.ROTATED_PLANAR,
                        observable_basis=PauliBasis.Z,
                        num_rounds=3,
                        basis_gates=["CZ", "H", "MZ", "RZ"],
                        parameters=CircuitParameters.from_sizes((3, 3))
                    )
                )

        """
        return CircuitParameters(sizes=Sizes(list(sizes)))

    @staticmethod
    def from_matrix_specification(
        param_l: int, param_m: int, m_A_powers: list[int], m_B_powers: list[int]
    ) -> CircuitParameters:
        """Use with Bivariate Bicycle codes.

        Args:
            param_l (int):
                The parameter `l`, used to construct a code of length `n = 2lm`.
            param_m (int):
                The parameter `m`, used to construct a code of length `n = 2lm`.
            m_A_powers (List[int]):
                The powers of the polynomial terms used to construct
                the matrix A. Each polynomial is of the form
                `x^a + y^b + y^c` so the sequence should specify
                `[a, b, c]`.
            m_B_powers (List[int]):
                The powers of the polynomial terms used to construct
                the matrix B. Each polynomial is of the form
                `y^a + x^b + x^c` so the sequence should specify
                `[a, b, c]`.

        Returns:
            CircuitParameters: object to pass to a circuit generation function.

        Examples:
            Creation and usage of parameters::

                circuit = client.generate_circuit(
                    types.QECExperimentDefinition(
                        experiment_type=types.QECExperimentType.QUANTUM_MEMORY,
                        code_type=types.QECECodeType.BIVARIATE_BICYCLE,
                        observable_basis=PauliBasis.Z,
                        num_rounds=6,
                        basis_gates=["CZ", "H", "MZ", "RZ"],
                        parameters=CircuitParameters.from_matrix_specification(
                            param_l=6,
                            param_m=6,
                            m_A_powers=[3, 1, 2],
                            m_B_powers=[3, 1, 2],
                        )
                    )
                )

        """
        return CircuitParameters(
            matrix_specifications=MatrixSpecifications(
                param_l=param_l,
                param_m=param_m,
                m_A_powers=m_A_powers,
                m_B_powers=m_B_powers
            )
        )

    def to_gql(self) -> dict[str, Any]:
        """Get a GraphQL-compatible dictionary
        representation of the object."""

        return {
            "sizes": self.sizes.to_gql() if self.sizes else None,
            "matrixSpecifications": (
                self.matrix_specifications.to_gql()
                if self.matrix_specifications
                else None
            ),
        }


@dataclass
class TypedData(ABC):  # pragma: nocover
    """Basic class for all typed data manipulation activities."""

    data_format: DataFormat
    """Format in which the data is stored. b8, 01, or csv."""

    content: Any
    """File name, on in-RAM array object."""

    data_width: int = -1
    """If data is serialisable, data_width supports b8 format parsing."""

    @abstractmethod
    def as_01_string(self) -> str:
        """Generate or return a 01 string for the stored data."""
        msg = "This is an abstract class"
        raise NotImplementedError(msg)

    @abstractmethod
    def as_b8_bytes(self) -> bytes:
        """Generate or return a b8 string for the stored data."""
        msg = "This is an abstract class"
        raise NotImplementedError(msg)

    @abstractmethod
    def as_numpy(self) -> npt.NDArray[np.uint8]:
        """Generate or return an numpy array for the stored data."""
        msg = "This is an abstract class"
        raise NotImplementedError(msg)

    def get_width(self) -> int:
        """Derives data width, infers from the data in 01 or CSV format."""
        if self.data_width != -1:
            return self.data_width
        if self.data_format != DataFormat.B8:
            self.data_width = self.as_numpy().shape[1]
            return self.data_width
        return -1


@dataclass
class RAMData(TypedData):
    """An wrapper around a collection stored in RAM."""

    content: list[list[int]] | npt.NDArray

    def get_width(self) -> int:
        """Infer the width of the string from the data."""
        if self.data_width != -1:
            return self.data_width
        if len(self.content) > 0:
            return len(self.content[0])
        return -1

    def as_01_string(self) -> str:
        string_builder = StringIO()
        write_01(self.content, string_builder)
        return string_builder.getvalue()

    def as_b8_bytes(self) -> bytes:
        bytes_builder = BytesIO()
        write_b8(self.content, bytes_builder)
        return bytes_builder.getvalue()

    def as_numpy(self) -> npt.NDArray[np.uint8]:
        return np.array(self.content, dtype=np.uint8)


@dataclass
class TypedDataFile(TypedData):
    """Data file with typed content."""

    content: Path

    def as_01_string(self) -> str:
        if self.data_format == DataFormat.F01:
            with Path.open(self.content, encoding="utf-8") as file:
                return file.read()
        elif self.data_format == DataFormat.CSV:
            with Path.open(self.content, encoding="utf-8") as file:
                string_builder = StringIO()
                write_01(read_csv(file), string_builder)
                return string_builder.getvalue()
        elif self.data_format == DataFormat.B8:
            with self.content.open("rb") as file:
                string_builder = StringIO()
                write_01(
                    read_b8(file, self.data_width), string_builder)
                return string_builder.getvalue()
        else:
            msg = "Type conversion is not implemented"
            raise NotImplementedError(msg)

    def as_b8_bytes(self) -> bytes:
        if self.data_format == DataFormat.B8:
            with self.content.open("rb") as file:
                return file.read()
        elif self.data_format == DataFormat.CSV:
            with Path.open(self.content, encoding="utf-8") as file:
                bytes_builder = BytesIO()
                write_b8(read_csv(file), bytes_builder)
                return bytes_builder.getvalue()
        elif self.data_format == DataFormat.F01:
            with Path.open(self.content, encoding="utf-8") as file:
                bytes_builder = BytesIO()
                write_b8(read_01(file), bytes_builder)
                return bytes_builder.getvalue()
        msg = f"Type conversion is not implemented for {self.data_format}"
        raise NotImplementedError(
            msg)

    def as_numpy(self) -> npt.NDArray[np.uint8]:
        if self.data_format == DataFormat.B8:
            with self.content.open("rb") as file:
                return np.array(
                    list(read_b8(file, self.data_width)), dtype=np.uint8)
        elif self.data_format == DataFormat.CSV:
            with Path.open(self.content, encoding="utf-8") as file:
                return np.array(list(read_csv(file)), dtype=np.uint8)
        elif self.data_format == DataFormat.F01:
            with Path.open(self.content, encoding="utf-8") as file:
                return np.array(list(read_01(file)), dtype=np.uint8)
        msg = f"Type conversion is not implemented for {self.data_format}"
        raise NotImplementedError(
            msg)


@dataclass
class TypedDataString(TypedData):
    """Typed data string."""

    content: DataString

    def as_01_string(self) -> str:
        if self.data_format == DataFormat.F01:
            return self.content.to_string()
        if self.data_format == DataFormat.B8:
            string_builder = StringIO()
            write_01(
                read_b8(BytesIO(self.content.data), self.data_width),
                string_builder,
            )
            return string_builder.getvalue()
        if self.data_format == DataFormat.CSV:
            string_builder = StringIO()
            write_01(
                read_csv(
                    TextIOWrapper(BytesIO(self.content.data), encoding="utf-8")
                ),
                string_builder,
            )
            return string_builder.getvalue()
        msg = "Type conversion is not implemented"
        raise NotImplementedError(msg)

    def as_b8_bytes(self) -> bytes:
        if self.data_format == DataFormat.B8:
            return self.content.data
        if self.data_format == DataFormat.CSV:
            bytes_builder = BytesIO()
            write_b8(
                read_csv(
                    TextIOWrapper(BytesIO(self.content.data), encoding="utf-8")
                ),
                bytes_builder
            )
            return bytes_builder.getvalue()
        if self.data_format == DataFormat.F01:
            bytes_builder = BytesIO()
            write_b8(
                read_01(
                    TextIOWrapper(BytesIO(self.content.data), encoding="utf-8")
                ),
                bytes_builder
            )
            return bytes_builder.getvalue()
        msg = "Type conversion is not implemented"
        raise NotImplementedError(msg)

    def as_numpy(self) -> npt.NDArray[np.uint8]:
        wrapper = TextIOWrapper(BytesIO(self.content.data), encoding="utf-8")

        if self.data_format == DataFormat.B8:
            return np.array(
                list(
                    read_b8(
                        stream=BytesIO(self.content.data),
                        width=self.data_width
                    )
                ),
                dtype=np.uint8,
            )
        if self.data_format == DataFormat.CSV:
            return np.array(
                list(read_csv(stream=wrapper)),
                dtype=np.uint8,
            )
        if self.data_format == DataFormat.F01:
            return np.array(
                list(read_01(stream=wrapper)),
                dtype=np.uint8,
            )
        msg = "Type conversion is not implemented"
        raise NotImplementedError(msg)

    @staticmethod
    def from_data(data: TypedData) -> TypedDataString:
        """
        Prepare a DataString of arbitrary table-like data,

        Args:
            data (QECDataType): File or table with the data to send.

        Returns:
            TypedDataString: DataString with associated data type.
        """
        if isinstance(data, TypedDataFile):
            return TypedDataString(
                data_format=data.data_format,
                content=DataString.from_file(data.content),
            )
        if isinstance(data, TypedDataString):
            return data
        if isinstance(data, RAMData):
            if data.data_format == DataFormat.F01:
                return TypedDataString(
                    data_format=data.data_format,
                    data_width=data.get_width(),
                    content=DataString(data.as_01_string()),
                )
            if data.data_format == DataFormat.CSV:
                string_builder = StringIO()
                write_csv(data.as_numpy(), string_builder)
                return TypedDataString(
                    data_format=data.data_format,
                    data_width=data.get_width(),
                    content=DataString(string_builder.getvalue()),
                )
            if data.data_format == DataFormat.B8:
                bytes_builder = BytesIO()
                write_b8(data.as_numpy(), bytes_builder)
                return TypedDataString(
                    data_format=data.data_format,
                    data_width=data.get_width(),
                    content=DataString(bytes_builder.getvalue()),
                )
        msg = f"Cannot extract binary data from {type(data)}, {data.data_format}."
        raise ValueError(
            msg)

    def as_data_string(self) -> str:
        """Gets a datastring for the object."""
        return str(self.content)


@dataclass
class BinaryDataType:
    """Any binary table, stored in a file, datastring, or RAM."""

    data: TypedData

    def __init__(
        self,
        data: Path | DataString | list[list[int]] | npt.NDArray,
        data_format: DataFormat = DataFormat.B8,
        data_width: int = -1,
    ):
        if isinstance(data, Path):
            self.data = TypedDataFile(
                data_format=data_format,
                content=data,
                data_width=data_width
            )
        elif isinstance(data, DataString):
            self.data = TypedDataString(
                data_format=data_format,
                content=data,
                data_width=data_width,
            )
        elif isinstance(data, (list, np.ndarray)):
            self.data = RAMData(
                data_format=data_format,
                content=np.array(data, dtype=np.uint8),
                data_width=data_width,
            )
        else:
            msg = f"Not supported argument of type {type(data)}"
            raise ValueError(msg)

    def as_data_string(self, data_format: DataFormat = DataFormat.B8) -> str:
        """Converts the object to a compact datastring."""
        if data_format == DataFormat.B8:
            return str(DataString(TypedDataString.from_data(self.data).as_b8_bytes()))
        if data_format == DataFormat.F01:
            return str(DataString(TypedDataString.from_data(self.data).as_01_string()))
        if data_format == DataFormat.CSV:
            string = StringIO()
            write_csv(
                TypedDataString.from_data(self.data).as_numpy(), string)
            return str(DataString(string.getvalue()))
        msg = f"Saving to data string is not implemented for {data_format}"
        raise NotImplementedError(
            msg)

    def as_01_string(self) -> str:
        """Represent content of the object as 01 string."""
        return self.data.as_01_string()

    def as_numpy(self) -> npt.NDArray[np.uint8]:
        """
        Represent the data as
        unsigned 1 byte/number (uint8) numpy table.
        """
        return self.data.as_numpy()

    def as_b8_bytes(self) -> bytes:
        """Represent underlying data as b8 bit-packed bytes."""
        return self.data.as_b8_bytes()

    def split(self, shards: int) -> list[BinaryDataType]:
        """Split a data file into multiple blocks based on the
        number of output files. Return objects of the same class.

        Args:
            shards (int): Number of approximately equal chunks.

        Returns:
            List[BinaryDataType]:
                List of chuck objects, of the same
                type as the calling object.

        Examples:
            Splitting the data file::

                object = Measurements(data_string, DataFormat.F01)
                parts = object.split(shards=shards)

        """
        # heavy lifting
        assert shards > 0, \
            f"Number of batches should be bigger than 0, but was {shards}."
        numpy_data = self.as_numpy()
        lines = numpy_data.shape[0]
        shard_size = ceil(lines / shards)
        if shards == 1:
            return [self]
        results = []
        cls = type(self)
        for i in range(shards):
            start, end = i * shard_size, (i + 1) * shard_size
            # this may produce empty objects, as in 6 / 4 -> [2, 2, 2, 0]
            results.append(
                cls(
                    data=numpy_data[start: end],
                    data_format=self.data.data_format,
                    data_width=self.data.get_width(),
                )
            )
        return results

    def to_batches(self, batch_size: int = 100_000) -> Iterable[BinaryDataType]:
        """Split a data file into batches of fixed size.
        Returns objects of the same class.

        Args:
            batch_size (int):
                Number of approximately equal chunks. Default is 100k.

        Returns:
            Iterable[BinaryDataType]:
                Iterable of batch objects, of the same
                type as the calling object.

        Examples:
            Splitting the data file::

                object = Measurements(data_string, DataFormat.F01)
                for batch in object.to_batches(1000):
                    ...

        """
        assert batch_size > 0, \
            f"Batch size should be bigger than 0, but was {batch_size}."
        numpy_data = self.as_numpy()
        lines = numpy_data.shape[0]
        cls = type(self)
        ptr = 0
        while ptr < lines:
            yield cls(
                data=numpy_data[ptr: ptr + batch_size],
                data_format=self.data.data_format,
                data_width=self.data.get_width(),
            )
            ptr += batch_size

    @classmethod
    def combine(cls, list_of_data: list[BinaryDataType]) -> BinaryDataType:
        """Stack multiple binary tables into one using numpy representation.

        Args:
            list_of_data (List[BinaryDataType]):
                list of objects, e.g. decoding predictions or measurements.

        Returns:
            BinaryDataType:
                an object of the class, which invoked the method.

        """
        data = np.vstack([d.as_numpy() for d in list_of_data])
        return cls(
            data=data,
            data_format=DataFormat.B8,
            data_width=data.shape[1],
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return np.allclose(self.as_numpy(), other.as_numpy())


class Measurements(BinaryDataType):
    """Binary table representing measurements."""

    @staticmethod
    def _deinstrument_leakage_circuit(circuit: str) -> str:
        """This method removes leakage heralding, noise, and
        detectors to make sure"""
        bits: list[int | None] = []
        measurement_bits_count = 0
        lines = circuit.splitlines()
        result_lines = []
        for line in lines:
            if line.strip() == "":
                continue
            if "HERALD_LEAKAGE_EVENT" in line:
                herald_gap = len(line.strip().split()) - 1
                bits.extend([None] * herald_gap)
                continue
            # measurement happened, adding new bits
            if line.strip()[0] == "M":
                measurements = len(line.strip().split()) - 1
                bits.extend(
                    range(measurement_bits_count, measurement_bits_count + measurements)
                )
                measurement_bits_count += measurements
            # detector/observable, update indices or skip if refers to leakage
            if ("DETECTOR" in line) or ("OBSERVABLE_INCLUDE" in line):
                skip_me = False
                # absolute values may become smaller, we should not re-remap!
                bits_from_latest_to_earliest = list(enumerate(bits))[::-1]
                for i, new_i in bits_from_latest_to_earliest:
                    # old relative indexing
                    neg_i = i - len(bits)
                    # this was a heralded leakage
                    if new_i is None:
                        if f"rec[{neg_i}]" in line:
                            # this line addresses leakage heralding
                            # we remove it
                            skip_me = True
                            break
                        # just do not update, if referred to leakage
                        continue
                    # new relative indexing
                    neg_new_i = new_i - measurement_bits_count
                    line = line.replace(f"rec[{neg_i}]", f"rec[{neg_new_i}]")
                if skip_me:
                    continue
            if ("RELAX" in line) or ("LEAKAGE" in line):
                continue
            result_lines.append(line)

        return "\n".join(result_lines)

    def to_detectors_and_observables(
        self,
        stim_circuit: str | stim.Circuit,
        sweep_bits: BinaryDataType | None = None,
    ) -> tuple[DetectionEvents, ObservableFlips]:
        """Converts measurements object into detectors and observables tuple.

        Args:
            stim_circuit (str | stim.Circuit):
                STIM circuit content.
            sweep_bits: (Optional[BinaryDataType]):
                If data was generated with initial sweep bits,
                they are needed to obtain correct detectors.

        Returns:
            Tuple[DetectionEvents, ObservableFlips]:
                Detector and observable data.
        """
        if isinstance(stim_circuit, stim.Circuit):
            circuit = stim_circuit
        else:
            try:
                circuit = stim.Circuit(stim_circuit)
            except ValueError:
                stim_circuit = self._deinstrument_leakage_circuit(stim_circuit)
                try:
                    circuit = stim.Circuit(stim_circuit)
                except ValueError as stim_ex:
                    msg = "Provided circuit is not a valid stim circuit."
                    raise ValueError(
                        msg
                    ) from stim_ex

        converter = circuit.compile_m2d_converter()
        sweeps = None
        if sweep_bits is not None:
            sweeps = sweep_bits.as_numpy().astype(np.bool_)
        dets, obs = converter.convert(
            measurements=self.as_numpy().astype(np.bool_),
            sweep_bits=sweeps,
            append_observables=False,
            separate_observables=True,
        )
        return (
            DetectionEvents(dets, self.data.data_format, circuit.num_detectors),
            ObservableFlips(obs, self.data.data_format, circuit.num_measurements),
        )


class DetectionEvents(BinaryDataType):
    """Binary table, representing detectors."""


class ObservableFlips(BinaryDataType):
    """Binary table, representing observables."""


class LeakageFlags(BinaryDataType):
    """Binary table, representing heralded leakage flags."""


@dataclass
class NoiseModel(ABC):
    """Generic noise model class."""

    ENDPOINT: ClassVar[APIEndpoints]
    """
    Each noise model has a corresponding service endpoint,
    which is responsible for applying this model to a circuit.
    """
    ENDPOINT_RESULT_FIELDNAME: ClassVar[str]
    """Endpoints name noise addition results differently."""


@dataclass
# pylint: disable=too-many-instance-attributes
class PhysicalNoiseModel(NoiseModel):
    """
    Physically-inspired noise model. This noise model
    is built on physical characteristics of qubits and
    gates.
    """

    ENDPOINT: ClassVar[APIEndpoints] = APIEndpoints.ADD_NOISE
    ENDPOINT_RESULT_FIELDNAME: ClassVar[str] = "addNoiseToStimCircuit"

    t_1: float
    """T1 time (relaxation from |1> to |0>), seconds."""

    t_2: float
    """T2 time (dephasing), seconds."""

    time_1_qubit_gate: float
    """Time to execute a 1-qubit gate."""

    time_2_qubit_gate: float
    """Time to execute a 2-qubit gate."""

    time_measurement: float
    """Time to measure a qubit."""

    time_reset: float
    """Time to reset a qubit."""

    p_1_qubit_gate_error: float
    """Probability of a flip while doing a 1-qubit gate."""

    p_2_qubit_gate_error: float
    """Probability of a flip while doing a 2-qubit gate."""

    p_reset_error: float
    """Probability of a flip while doing a reset."""

    p_meas_qubit_error: float
    """Probability of incorrect measurement."""

    p_readout_flip: float
    """Probability of a flip while measuring a qubit."""

    @staticmethod
    def get_floor_superconducting_noise():
        """Return minimum supercondicting noise."""
        return PhysicalNoiseModel(
            t_1=20e-6,
            t_2=30e-6,
            time_1_qubit_gate=25e-9,
            time_2_qubit_gate=34e-9,
            time_measurement=500e-9,
            time_reset=160e-9,
            p_1_qubit_gate_error=0.312e-3,
            p_2_qubit_gate_error=0.425e-3,
            p_reset_error=1.99e-3,
            p_meas_qubit_error=6.17e-3,
            p_readout_flip=0.0
        )

    @staticmethod
    def get_superconducting_noise():
        """Generic superconducting noise model, derived from Google's
        https://www.nature.com/articles/s41586-022-05434-1 experiment."""
        return PhysicalNoiseModel(
            t_1=20e-6,
            t_2=30e-6,
            time_1_qubit_gate=25e-9,
            time_2_qubit_gate=34e-9,
            time_measurement=500e-9,
            time_reset=160e-9,
            p_1_qubit_gate_error=0.001,
            p_2_qubit_gate_error=0.01,
            p_reset_error=0.01,
            p_meas_qubit_error=0.01,
            p_readout_flip=0.01
        )

    @staticmethod
    def get_ion_trap_noise():
        """Generic ion trap noise model, derived from
        IonQ Aria public device specification
        https://ionq.com/quantum-systems/aria."""

        return PhysicalNoiseModel(
            t_1=100.0,
            t_2=1.0,
            time_1_qubit_gate=5e-6,
            time_2_qubit_gate=2.5e-5,
            time_measurement=1.2e-4,
            time_reset=1e-5,
            p_1_qubit_gate_error=6e-4,
            p_2_qubit_gate_error=6e-3,
            p_reset_error=1e-5,
            p_meas_qubit_error=39e-4,
            p_readout_flip=3e-3
        )


@dataclass
class SI1000NoiseModel(NoiseModel):
    """
    The SI 1000
    (Superconducting Inspired with 1000 nanosecond cycle)
    noise model, with leakage.
    See https://arxiv.org/pdf/2312.04522v1 (Appendix A) for details.
    """

    ENDPOINT: ClassVar[APIEndpoints] = APIEndpoints.ADD_SI1000_NOISE
    ENDPOINT_RESULT_FIELDNAME: ClassVar[str] = "addSi1000NoiseToStimCircuit"

    p: float
    """Probability of a Pauli error."""

    p_l: float
    """Probability of a leakage error."""


class DecodingResult:
    """Results of decoding"""

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        fails: int,
        shots: int,
        times: list[float] | None = None,
        counts: list[int] | None = None,
        predictions_format: DataFormat = DataFormat.F01,
        predictionsFile: dict[str, str | None] | None = None,
    ):
        """
        Args:
            fails (int): number of times decoder failed
            shots (int): number of shots performed
            times (List[float]): CPU time for decoding workers
            counts (List[int]): number of shots each worker processed
            predictions_format (DataFormat): format of predictions
            predictionsFile (Optional[Dict[str, Optional[str]]]):
                definition of filename or datastring with decoding predictions.
        """
        self.fails = fails
        self.shots = shots
        if times is None:
            times = []
        if counts is None:
            counts = [shots]
        self.times = times
        self.counts = counts
        self.predictions_format = predictions_format
        self.predictions_file = predictionsFile
        if self.predictions_file is not None:
            string = self.predictions_file.get("uid")
            if string is None:
                self.predictions = None
            else:
                self.predictions = BinaryDataType(
                    data=DataString.from_data_string(string),
                    data_format=self.predictions_format,
                )
        else:
            self.predictions = None

    def get_logical_error_probability(self) -> float:
        """Logical error probability, fails / shots."""
        if self.shots <= 0:
            return float("inf")
        return self.fails / self.shots

    def get_standard_deviation(self) -> float:
        """
        Return standard deviation of result
        for a given number of shots and decoder fails."""
        if self.shots <= 0:
            return float("inf")
        hits = self.shots - self.fails
        return (self.fails * hits / self.shots ** 3) ** .5

    @classmethod
    def combine(cls, results: Iterable[DecodingResult]) -> DecodingResult:
        """Concatenate decoding results from multiple experiments.
        Predictions are also merged.

        Args:
            results (Iterable[DecodingResult]):
                collection of decoding results. E.g. from batch decoding.

        Returns:
            DecodingResult: combined results, including predictions.
        """
        fails = 0
        shots = 0
        times: list[float] = []
        counts: list[int] = []
        predictions = []
        for result in results:
            fails += result.fails
            shots += result.shots
            times.extend(result.times)
            counts.extend(result.counts or [result.shots])
            # pylint: disable=protected-access
            if result.predictions is not None:
                predictions.append(result.predictions)
        total_result = cls(
            fails=fails,
            shots=shots,
            times=times,
            counts=counts,
        )
        if len(predictions) > 0:
            total_result.predictions = BinaryDataType.combine(predictions)
        return total_result

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, DecodingResult)
            and self.counts == other.counts
            and self.fails == other.fails
            and self.shots == other.shots
            and self.predictions == other.predictions
        )

    def __str__(self) -> str:
        return (
            f"DecodingResult(shots={self.shots}, "
            f"fails={self.fails}, "
            f"LEP={self.get_logical_error_probability():.5f}"
            f" ± {self.get_standard_deviation():.5f})"
        )

    def __repr__(self) -> str:
        return self.__str__()


@dataclass
class Decoder:
    """Common settings for decoding."""

    decoder_type: DecoderType
    """Decoding algorithm."""

    use_experimental_graph: bool = False
    """If set to True, data will be split into halves.
    One half of the data will be used for training experimental
    decoder error graph, and the other half will be decoded with
    this graph. And vice versa."""

    parallel_jobs: int = 1
    """Server will perform data-parallel decoding, if more cores
    are requested."""

    parameters: dict[str, Any] = field(default_factory=dict)
    """Decoders may have individual settings. They may be passed as
    dictionary items."""


@dataclass
class QubitCoordinateToDetectorMapping:
    """Mapping of detectors to qubit coordinates."""

    detector_map: dict[tuple[float, ...], list[int]]
    """Keys are detector coordinates, values are a list of
    all detectors IDs, associated with these coordinates."""



from enum import Enum, auto
from typing import Protocol

import numpy as np
import numpy.typing as npt


class GradientFitDiscretisationGenerator(Protocol):
    def __call__(
        self, a: float, b: float, c: float, num_points: int, degree: int, /
    ) -> npt.NDArray[np.floating]:
        """
        Find a good set of points to estimate the gradient at point ``c`` by fitting a
        polynomial of degree ``degree`` on the interval ``[a, b]``.

        Functions implementing that interface find ``num_points`` unique points ``x``
        between ``a`` and ``b`` that will be used to fit a polynomial of degree
        ``degree``.

        Different implementation might choose to minimise different quantities. For
        example:

        - minimise code complexity, by simply returning linear (or log-spaced) points,
        - minimise the variance of the resulting gradient estimation, for example using
          a D-optimal design (sees
          https://en.wikipedia.org/wiki/Optimal_experimental_design).

        Args:
            a (float): lower bound of the interval in which fitting points should be
                computed.
            b (float): upper bound of the interval in which fitting points should be
                computed.
            c (float): point at which the gradient will be estimated.
            num_points (int): number of points to return within ``[a, b]``.
            degree (int): degree of the polynomial that will be used to fit the values
                computed at each of the points returned by this function.

        Returns:
            a sorted array of ``num_points`` points within ``[a, b]`` and without duplicates
            that should be used to evaluate the function to fit with a degree ``degree``
            polynomial.

        Raises:
            ValueError: if ``a < c < b`` is not verified.
        """
        ...


def _check_interval(a: float, b: float, c: float) -> None:
    if not a < c < b:
        msg = f"Expected {a=} < {c=} < {b=}"
        raise ValueError(msg)


def get_linear_points(
    a: float, b: float, c: float, num_points: int, _: int
) -> npt.NDArray[np.floating]:
    """Returns ``num_points`` linearly spaced between ``a`` and ``b``."""
    _check_interval(a, b, c)
    return np.linspace(a, b, num_points)


def get_logarithmic_points(
    a: float, b: float, c: float, num_points: int, _: int
) -> npt.NDArray[np.floating]:
    """Returns ``num_points`` logarithmically spaced between ``a`` and ``b``."""
    _check_interval(a, b, c)
    if a <= 0:
        msg = (
            "Cannot get logarithmically-spaced points for negative values. "
            f"Got [{a}, {b}]."
        )
        raise ValueError(msg)
    return np.logspace(np.log10(a), np.log10(b), num_points, base=10)


class DiscretisationStrategy(Enum):
    """Strategy to use to generate discretisation point for fitting a noisy function
    with a polynomial."""

    LINEAR = auto()
    """Linearly spaced points between the discretisation space boundaries."""
    LOGARITHMIC = auto()
    """Logarithmically spaced points between the discretisation space boundaries."""

    def __call__(
        self, a: float, b: float, c: float, num_points: int, degree: int, /
    ) -> npt.NDArray[np.floating]:
        match self:
            case DiscretisationStrategy.LINEAR:
                return get_linear_points(a, b, c, num_points, degree)
            case DiscretisationStrategy.LOGARITHMIC:
                return get_logarithmic_points(a, b, c, num_points, degree)
            case _:
                msg = f"Discretisation {self} is not implemented yet."
                raise NotImplementedError(msg)
